@@ -478,6 +478,54 @@ quick_lounge_util__get_unique_uri (QuickLounge *quick_lounge)
 /* -- */
 
 
+gboolean
+quick_lounge_add_launcher (QuickLounge  *quick_lounge,
+			   GKeyFile     *desktop_entry,
+			   int           position)
+{
+	gboolean  launcher_added = FALSE;
+	GError   *error = NULL;
+	char     *data;
+	gsize     length;
+	char     *uri;
+	GFile    *file;
+	
+	if (desktop_entry == NULL)
+		return FALSE;
+	
+	data = g_key_file_to_data (desktop_entry, &length, &error);
+	if (data == NULL) {
+		g_key_file_free (desktop_entry);
+		_gtk_error_dialog_run (NULL,
+				       ERROR_FORMAT,
+				       _("Could not create the new launcher"),
+				       error->message);
+		g_clear_error (&error);
+		return FALSE;
+	} 
+	
+	uri = quick_lounge_util__get_unique_uri (quick_lounge);
+	file = g_file_new_for_uri (uri);
+	if (! g_write_file (file, FALSE, G_FILE_CREATE_NONE, data, length, NULL, &error)) { 
+		_gtk_error_dialog_run (NULL,
+				       ERROR_FORMAT,
+				       _("Could not create the new launcher"),
+				       error->message);
+		g_clear_error (&error);
+	}
+	else {
+		quick_box_add_button (quick_lounge->quick_box, uri, position);
+		launcher_added = TRUE;
+	}
+	
+	g_object_unref (file);
+	g_free (uri);
+	g_free (data);
+	
+	return launcher_added;
+}
+
+
 static void
 new_desktop_entry_dialog_response_cb (GtkDialog *dialog,
 				      int        response,
@@ -486,10 +534,6 @@ new_desktop_entry_dialog_response_cb (GtkDialog *dialog,
 	EditItemData *edit_data = user_data;
 	GKeyFile     *desktop_entry;
 	GError       *error = NULL;
-	char         *data;
-	gsize         length;
-	char         *uri;
-	GFile        *file;
 		
 	switch (response) {
 	case GTK_RESPONSE_HELP:
@@ -517,40 +561,12 @@ new_desktop_entry_dialog_response_cb (GtkDialog *dialog,
 		return;
 	}
 	
-	if (desktop_entry == NULL)
-		return;
-	
-	data = g_key_file_to_data (desktop_entry, &length, &error);
-	if (data == NULL) {
-		g_key_file_free (desktop_entry);
-		_gtk_error_dialog_run (NULL,
-				       ERROR_FORMAT,
-				       _("Could not create the new launcher"),
-				       error->message);
-		g_clear_error (&error);
-		return;
-	} 
-	
-	uri = quick_lounge_util__get_unique_uri (edit_data->quick_lounge);
-	file = g_file_new_for_uri (uri);
-	if (! g_write_file (file, FALSE, G_FILE_CREATE_NONE, data, length, NULL, &error)) { 
-		_gtk_error_dialog_run (NULL,
-				       ERROR_FORMAT,
-				       _("Could not create the new launcher"),
-				       error->message);
-		g_clear_error (&error);
-	}
-	else {
-		quick_box_add_button (edit_data->quick_lounge->quick_box, uri, edit_data->position + 1);
+	if (quick_lounge_add_launcher (edit_data->quick_lounge, desktop_entry, edit_data->position + 1)) {
 		quick_lounge_save_order (edit_data->quick_lounge);
 		dlg_properties_update (edit_data->quick_lounge->prop_dialog);
-
 		gtk_widget_destroy (GTK_WIDGET (dialog));
 	}
 	
-	g_object_unref (file);
-	g_free (uri);
-	g_free (data);
 	g_key_file_free (desktop_entry);
 }
 
@@ -784,8 +800,6 @@ open_file_list_with_launcher (QuickLounge *quick_lounge,
 	desktop_entry = g_object_get_data (G_OBJECT (button), "desktop_entry");
 	if (desktop_entry != NULL)
 		_g_desktop_entry_launch (desktop_entry, button, list);
-		
-	path_list_free (list);
 }
 
 
@@ -799,11 +813,13 @@ drag_data_received  (GtkWidget          *widget,
 		     guint               time,
 		     gpointer            extra_data)
 {
-	QuickLounge *quick_lounge = extra_data;
-	QuickBox    *quick_box = QUICK_BOX (widget);
-        GList       *list, *scan;
-	int          pos;
-
+	QuickLounge  *quick_lounge = extra_data;
+	QuickBox     *quick_box = QUICK_BOX (widget);
+        GList        *list, *scan;
+	int           pos;
+	QuickDropPos  drop_position;
+	gboolean      launcher_list_changed = FALSE;
+	
         if (! ((data->length >= 0) && (data->format == 8))) {
                 gtk_drag_finish (context, FALSE, FALSE, time);
                 return;
@@ -811,52 +827,82 @@ drag_data_received  (GtkWidget          *widget,
 
 	gtk_drag_finish (context, TRUE, FALSE, time);
 
-	pos = quick_box_get_pointer_position (quick_box, x, y);
+	pos = quick_box_get_drop_position (quick_box, x, y, &drop_position);
 	list = get_file_list_from_url_list ((char *) data->data);
 
-	for (scan = list; scan; scan = scan->next) {
-		char      *src_uri = scan->data;
-		char      *uri;
-		GtkWidget *button;
-
-		if (! is_desktop_file (src_uri)) {
+	if ((drop_position == QUICK_DROP_POSITION_INTO_OR_BEFORE)
+	    || (drop_position == QUICK_DROP_POSITION_INTO_OR_AFTER))
+	{
+		if (! is_desktop_file (list->data)) { 
 			open_file_list_with_launcher (quick_lounge, pos, list);
+			path_list_free (list);
 			return;
 		}
-
-		uri = g_build_filename (quick_lounge->location,
-					file_name_from_path (src_uri),
-					NULL);
-
-		if (strcmp (src_uri, uri) == 0) {
-			button = quick_box_get_child_from_uri (quick_box, uri);
-
-			if (button != NULL) {
-				int button_pos;
-				button_pos = quick_box_get_child_position (quick_box, button);
-				if (button_pos != -1)
-					quick_box_reorder_child (quick_box, button, pos);
-			}
-
-			g_free (uri);
-			continue;
-		}
-
-		g_free (uri);
-		uri = quick_lounge_util__get_unique_uri (quick_lounge);
-
-		if (! file_copy (src_uri, uri)) {
-			g_free (uri);
-			continue;
-		}
-
-		quick_box_add_button (quick_box, uri, ++pos);
-		g_free (uri);
 	}
-	path_list_free (list);
 
-	quick_lounge_save_order (quick_lounge);
-	dlg_properties_update (quick_lounge->prop_dialog);
+	if ((drop_position == QUICK_DROP_POSITION_AFTER)
+	    || (drop_position == QUICK_DROP_POSITION_INTO_OR_AFTER))
+	{
+		pos++;
+	}
+	
+	for (scan = list; scan; scan = scan->next) {
+		char *src_uri = scan->data;
+			
+		if (is_desktop_file (src_uri)) { 
+			char      *uri;
+			GtkWidget *button;
+
+			uri = g_build_filename (quick_lounge->location,
+						file_name_from_path (src_uri),
+						NULL);
+	
+			if (strcmp (src_uri, uri) == 0) {
+				button = quick_box_get_child_from_uri (quick_box, uri);
+	
+				if (button != NULL) {
+					int button_pos;
+					button_pos = quick_box_get_child_position (quick_box, button);
+					if (button_pos != -1) {
+						quick_box_reorder_child (quick_box, button, pos);
+						launcher_list_changed = TRUE;
+					}
+				}
+	
+				g_free (uri);
+				continue;
+			}
+	
+			g_free (uri);
+			uri = quick_lounge_util__get_unique_uri (quick_lounge);
+	
+			if (! file_copy (src_uri, uri)) {
+				g_free (uri);
+				continue;
+			}
+	
+			quick_box_add_button (quick_box, uri, pos++);
+			launcher_list_changed = TRUE;
+			
+			g_free (uri);
+		}
+		else {
+			GKeyFile *desktop_entry;
+				
+			desktop_entry = _g_desktop_entry_new_for_uri (src_uri);
+			if (quick_lounge_add_launcher (quick_lounge, desktop_entry, pos)) 
+				launcher_list_changed = TRUE;
+						
+			g_key_file_free (desktop_entry);
+		}
+	}
+	
+	if (launcher_list_changed) {
+		quick_lounge_save_order (quick_lounge);
+		dlg_properties_update (quick_lounge->prop_dialog);
+	}
+
+	path_list_free (list);
 }
 
 
